import unittest

from django_schema_sprout.utils.singleton_class import SingletonArgs


class SingletonMetaTestCase(unittest.TestCase):
    def setUp(self):
        class DummyClass(metaclass=SingletonArgs):
            def __init__(self, arg1, arg2):
                pass

        class DummyClassWithoutInit(metaclass=SingletonArgs):
            pass

        self.DummyClass = DummyClass
        self.DummyClassWithoutInit = DummyClassWithoutInit

    def test_two_initalizations_with_same_args(self):
        # Initialize two instances of DummyClass with the same arguments
        dummy_instance_1 = self.DummyClass("test", "testButDiiferent")
        dummy_instance_2 = self.DummyClass("test", "testButDiiferent")

        # Assert that the two instances are the same
        self.assertEqual(id(dummy_instance_1), id(dummy_instance_2))

    def test_two_initalizations_with_different_args(self):
        # Initialize two instances of DummyClass with the same arguments
        dummy_instance_1 = self.DummyClass("test", "testButDiiferent")
        dummy_instance_2 = self.DummyClass("test", "testButThisTimeDiiferent")

        # Assert that the two instances are not the same
        self.assertNotEqual(id(dummy_instance_1), id(dummy_instance_2))

    def test_two_initalizations_with_no_args(self):
        # Initialize two instances of DummyClass with no arguments
        dummy_instance_1 = self.DummyClassWithoutInit()
        dummy_instance_2 = self.DummyClassWithoutInit()

        # Assert that the two instances are the same
        self.assertEqual(id(dummy_instance_1), id(dummy_instance_2))



from django.test import TestCase
from django.test.utils import isolate_apps
from django.db import models
from rest_framework import serializers
from rest_framework import viewsets
from django_schema_sprout.utils.dynamic_view import create_view


class TestDynamicView(TestCase):
    @isolate_apps("test")
    def test_create_view(self):
        class TestModel(models.Model):
            class Meta:
                app_label = "test"

            pass

        # Create a test serializer
        class TestSerializer(serializers.ModelSerializer):
            class Meta:
                model = TestModel
                fields = "__all__"

        # Create a view class using the create_view function
        view_class = create_view(TestModel, TestSerializer)

        # Assert that the view class is a subclass of the appropriate viewset class
        self.assertTrue(issubclass(view_class, viewsets.ModelViewSet))

        self.assertEqual(view_class.serializer_class, TestSerializer)

    @isolate_apps("test")
    def test_create_view_readonly(self):
        class TestModel(models.Model):
            class Meta:
                app_label = "test"

            pass

        # Create a test serializer
        class TestSerializer(serializers.ModelSerializer):
            class Meta:
                model = TestModel
                fields = "__all__"

        readonly_view_class = create_view(TestModel, TestSerializer, readonly=True)

        self.assertTrue(issubclass(readonly_view_class, viewsets.ReadOnlyModelViewSet))

        self.assertEqual(readonly_view_class.serializer_class, TestSerializer)



from unittest.mock import Mock

from django.test import TestCase

from django_schema_sprout.utils.db_inspector import get_inspector
from django_schema_sprout.db_inspectors.postgres_inspector import PostgresDBInspect


class DBInspectorTestCase(TestCase):
    def test_get_inspector_postgresql(self):
        connection = Mock()
        connection.vendor = "postgresql"

        # Call the get_inspector function
        inspector = get_inspector(connection)

        # Assert that the returned inspector is an instance of PostgresDBInspect
        self.assertIsInstance(inspector, PostgresDBInspect)

    def test_get_inspector_unsupported_database(self):
        connection = Mock()
        connection.vendor = "mysql"

        # Call the get_inspector function and assert raise of NotImplementedError
        with self.assertRaises(NotImplementedError):
            get_inspector(connection)



from django_schema_sprout.db_inspectors.postgres_inspector import PostgresDBInspect


def get_inspector(connection):
    if connection.vendor == "postgresql":
        return PostgresDBInspect(connection)
    else:
        raise NotImplementedError(f"Database {connection.vendor} not supported")



from django.db import models
from rest_framework import serializers


def create_serializer(model: models.Model) -> type:
    fields = [field.name for field in model._meta.fields]

    meta_class = type(
        "Meta",
        (),
        {"model": model, "fields": fields},
    )

    return type(
        f"{model.__name__}Serializer",
        (serializers.ModelSerializer,),
        {"Meta": meta_class},
    )



from django.db import models

from django_schema_sprout.apps import SchemaSproutAppConfig
from django_schema_sprout.models import DynamicDbModel


def create_model(
    database: str, table_name: str, attrs: dict, nspname: str
) -> models.Model:
    class Meta(DynamicDbModel.Meta):
        db_table = f'"{nspname}"."{table_name}"'
        _schema_sprout_db = database
        managed = False
        app_label = SchemaSproutAppConfig.name

    attrs["Meta"] = Meta
    attrs["__module__"] = f"{SchemaSproutAppConfig.name}.models"
    model = type(f"{nspname}_{table_name}", (DynamicDbModel,), attrs)
    model.Meta._schema_sprout_db = database

    return model



from django.test import TestCase
from django.db import models
from django_schema_sprout.apps import SchemaSproutAppConfig
from django_schema_sprout.models import DynamicDbModel
from django_schema_sprout.utils.dynamic_model import create_model


class DynamicModelTestCase(TestCase):
    def test_create_model(self):
        # Define attributes for the model
        attrs = {
            "field1": models.CharField(max_length=100),
            "field2": models.IntegerField(),
        }

        # Create a model using the create_model function
        model = create_model("my_database", "my_table", attrs, "my_schema")

        # Assert that the model is a subclass of DynamicDbModel
        self.assertTrue(issubclass(model, DynamicDbModel))

        # Assert that the model has the correct Meta attributes

        self.assertEqual(model._meta.db_table, '"my_schema"."my_table"')
        self.assertFalse(model._meta.managed)
        self.assertEqual(model.Meta._schema_sprout_db, "my_database")
        self.assertEqual(model.Meta.app_label, SchemaSproutAppConfig.name)

        # Assert that the model has the correct fields
        self.assertIsInstance(model.field1.field, models.CharField)
        self.assertIsInstance(model.field2.field, models.IntegerField)

        # Assert that the model has the correct name
        self.assertEqual(model.__name__, "my_schema_my_table")



import logging

from django.db import connections
from django.db.utils import ProgrammingError
from rest_framework.routers import DefaultRouter

from django_schema_sprout.utils.db_inspector import get_inspector
from django_schema_sprout.utils.dynamic_view import create_view
from django_schema_sprout.utils.dynamic_model import create_model
from django_schema_sprout.utils.dynamic_serializer import create_serializer
from django_schema_sprout.utils.singleton_class import SingletonArgs


class SchemaSprout(metaclass=SingletonArgs):

    def __init__(self, database: str):
        self.database = database
        self.connection = connections[self.database]
        self.db_inspector = get_inspector(self.connection)

        self.views = dict()
        self.serializers = dict()
        self.models = dict()

        self.router = DefaultRouter()

    def create_models(self, readonly: bool = False):
        # Create models only for tables
        types = {"t"}
        with self.connection.cursor() as cursor:
            tables_info = self.db_inspector.get_table_list(cursor)

        for table in tables_info:
            if table.type not in types:
                continue
            table_name = table.name
            table_nspname = table.nspname
            self.create_model(table_name, table_nspname, readonly)

    def create_model(self, table_name, table_nspname, readonly: bool = False):
        nsp_name = f"{table_nspname}_{table_name}"
        if (
            nsp_name in self.models
            and nsp_name in self.serializers
            and nsp_name in self.views
        ):
            return

        with self.connection.cursor() as cursor:
            try:
                attrs = self.db_inspector.get_attributes(
                    cursor, table_name, table_nspname
                )

                model = self.models.get(
                    nsp_name,
                    create_model(self.database, table_name, attrs, table_nspname),
                )
                self.models[nsp_name] = model

                serializer = self.serializers.get(nsp_name, create_serializer(model))
                self.serializers[nsp_name] = serializer

                view = create_view(model, serializer, readonly)
                self.views[nsp_name] = view

                _table_nspname = table_nspname.lower().replace("_", "-")
                _table_name = table_name.lower().replace("_", "-")

                self.router.register(f"{_table_nspname}/{_table_name}", view)

            except ProgrammingError as err:
                logging.warning(
                    f"Couldn't get table description for table {table_name}. "
                    f"Reason: {err}"
                )



from collections import namedtuple


from django.db import models
from django.db.models import Index
from django.db.backends.base.introspection import TableInfo as BaseTableInfo
from django.db.backends.base.introspection import FieldInfo as BaseFieldInfo
from django.db.backends.postgresql.introspection import DatabaseIntrospection
from django.core.management.commands.inspectdb import Command


TableInfo = namedtuple(
    "TableInfo",
    BaseTableInfo._fields
    + (
        "comment",
        "nspname",
    ),
)
FieldInfo = namedtuple("FieldInfo", BaseFieldInfo._fields + ("is_autofield", "comment"))


class PostgresDBInspect(DatabaseIntrospection):
    """
    A class for inspecting the PostgreSQL database schema.

    This class provides methods to retrieve information about tables, relations,
    constraints, primary keys, and table descriptions in a PostgreSQL database.
    """

    def get_table_list(self, cursor):
        """
        Return a list of table and view names in
        the current database in order of dependency.
        """
        cursor.execute(
            """
                with recursive fk_tree as (
                -- All tables not referencing anything else except self references
                SELECT
                    c.oid as reloid,
                    c.relname as table_name,
                    n.nspname as schema_name,
                    CASE
                        WHEN c.relispartition THEN 'p'
                        WHEN c.relkind IN ('m', 'v') THEN 'v'
                        ELSE 't'
                    END AS relcase,
                    obj_description(c.oid, 'pg_class') AS obj_desc,
                    null::text COLLATE "C" as referenced_table_name,
                    null::text COLLATE "C" as referenced_schema_name,
                    1 AS level
                FROM pg_catalog.pg_class c
                LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
                WHERE c.relkind IN ('f', 'm', 'p', 'r', 'v')
                AND n.nspname NOT IN ('pg_catalog', 'pg_toast')
                AND n.nspname = ANY(current_schemas(false))
                AND NOT EXISTS (
                    SELECT *
                    FROM pg_constraint
                    WHERE contype = 'f'
                    -- self-referencing-check
                    AND conrelid = c.oid AND conrelid != confrelid
                )
                UNION ALL
                SELECT
                    ref.oid,
                    ref.relname,
                    rs.nspname,
                    CASE
                        WHEN ref.relispartition THEN 'p'
                        WHEN ref.relkind IN ('m', 'v') THEN 'v'
                        ELSE 't'
                    END,
                    obj_description(c.oid, 'pg_class'),
                    p.table_name,
                    p.schema_name,
                    p.level + 1
                FROM pg_class ref
                    JOIN pg_namespace rs ON rs.oid = ref.relnamespace
                    JOIN pg_constraint c ON c.contype = 'f' AND c.conrelid = ref.oid
                    JOIN fk_tree p ON p.reloid = c.confrelid
                WHERE ref.relkind IN ('f', 'm', 'p', 'r', 'v')
                -- do not enter to tables referencing themselves.
                AND ref.oid != p.reloid
                )
                SELECT
                    table_name,
                    relcase,
                    obj_desc,
                    schema_name
                FROM fk_tree
                ORDER BY "level" asc, schema_name, table_name;
            """
        )

        return [
            TableInfo(*row)
            for row in cursor.fetchall()
            if row[0] not in self.ignored_tables
        ]

    def get_relations(self, cursor, table_name, nspname):
        """
        Return a dictionary of {field_name: (field_name_other_table, other_table)}
        representing all foreign keys in the given table.
        """
        cursor.execute(
            """
            SELECT a1.attname, c2.relname, a2.attname, n2.nspname
            FROM pg_constraint con
            LEFT JOIN pg_class c1 ON con.conrelid = c1.oid
            LEFT JOIN pg_class c2 ON con.confrelid = c2.oid
            LEFT JOIN
                pg_attribute a1 ON c1.oid = a1.attrelid AND a1.attnum = con.conkey[1]
            LEFT JOIN
                pg_attribute a2 ON c2.oid = a2.attrelid AND a2.attnum = con.confkey[1]
            LEFT JOIN pg_catalog.pg_namespace n1 ON n1.oid = c1.relnamespace
            LEFT JOIN pg_catalog.pg_namespace n2 ON n2.oid = c2.relnamespace
            WHERE
                c1.relname = %s AND
                con.contype = 'f' AND
                c1.relnamespace = c2.relnamespace AND
                n1.nspname = ANY(current_schemas(false)) AND
                n1.nspname = %s
        """,
            [table_name, nspname],
        )
        return {row[0]: (row[2], row[1], row[3]) for row in cursor.fetchall()}

    def get_constraints(self, cursor, table_name, nspname):
        """
        Retrieve any constraints or keys (unique, pk, fk, check, index) across
        one or more column in given namespace.
        Also retrieve the definition of expression-based indexes.
        """
        constraints = {}
        # Loop over the key table, collecting things as constraints. The column
        # array must return column names in the same order in which they were
        # created.
        cursor.execute(
            """
            SELECT
                c.conname,
                array(
                    SELECT attname
                    FROM unnest(c.conkey) WITH ORDINALITY cols(colid, arridx)
                    JOIN pg_attribute AS ca ON cols.colid = ca.attnum
                    WHERE ca.attrelid = c.conrelid
                    ORDER BY cols.arridx
                ),
                c.contype,
                (SELECT fkc.relname || '.' || fka.attname
                FROM pg_attribute AS fka
                JOIN pg_class AS fkc ON fka.attrelid = fkc.oid
                WHERE fka.attrelid = c.confrelid AND fka.attnum = c.confkey[1]),
                cl.reloptions
            FROM pg_constraint AS c
            JOIN pg_class AS cl ON c.conrelid = cl.oid
            LEFT JOIN pg_catalog.pg_namespace n ON n.oid = cl.relnamespace
            WHERE cl.relname = %s
            AND n.nspname = ANY(current_schemas(false))
            AND n.nspname = %s
        """,
            [table_name, nspname],
        )
        for constraint, columns, kind, used_cols, options in cursor.fetchall():
            constraints[constraint] = {
                "columns": columns,
                "primary_key": kind == "p",
                "unique": kind in ["p", "u"],
                "foreign_key": tuple(used_cols.split(".", 1)) if kind == "f" else None,
                "check": kind == "c",
                "index": False,
                "definition": None,
                "options": options,
            }
        # Now get indexes
        cursor.execute(
            """
            SELECT
                indexname,
                array_agg(attname ORDER BY arridx),
                indisunique,
                indisprimary,
                array_agg(ordering ORDER BY arridx),
                amname,
                exprdef,
                s2.attoptions
            FROM (
                SELECT
                    c2.relname as indexname, idx.*, attr.attname, am.amname,
                    CASE
                        WHEN idx.indexprs IS NOT NULL THEN
                            pg_get_indexdef(idx.indexrelid)
                    END AS exprdef,
                    CASE am.amname
                        WHEN %s THEN
                            CASE (option & 1)
                                WHEN 1 THEN 'DESC' ELSE 'ASC'
                            END
                    END as ordering,
                    c2.reloptions as attoptions
                FROM (
                    SELECT *
                    FROM
                        pg_index i,
                        unnest(i.indkey, i.indoption)
                            WITH ORDINALITY koi(key, option, arridx)
                ) idx
                LEFT JOIN pg_class c ON idx.indrelid = c.oid
                LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
                LEFT JOIN pg_class c2 ON idx.indexrelid = c2.oid
                LEFT JOIN pg_am am ON c2.relam = am.oid
                LEFT JOIN
                    pg_attribute attr ON attr.attrelid = c.oid AND attr.attnum = idx.key
                WHERE c.relname = %s
                AND n.nspname = ANY(current_schemas(false))
                AND n.nspname = %s
            ) s2
            GROUP BY indexname, indisunique, indisprimary, amname, exprdef, attoptions;
        """,
            [self.index_default_access_method, table_name, nspname],
        )
        for (
            index,
            columns,
            unique,
            primary,
            orders,
            type_,
            definition,
            options,
        ) in cursor.fetchall():
            if index not in constraints:
                basic_index = (
                    type_ == self.index_default_access_method
                    and
                    # '_btree' references
                    # django.contrib.postgres.indexes.BTreeIndex.suffix.
                    not index.endswith("_btree")
                    and options is None
                )
                constraints[index] = {
                    "columns": columns if columns != [None] else [],
                    "orders": orders if orders != [None] else [],
                    "primary_key": primary,
                    "unique": unique,
                    "foreign_key": None,
                    "check": False,
                    "index": True,
                    "type": Index.suffix if basic_index else type_,
                    "definition": definition,
                    "options": options,
                }
        return constraints

    def get_primary_key_column(self, cursor, table_name, nspname):
        """
        Return the name of the primary key column for the given table.
        """
        columns = self.get_primary_key_columns(cursor, table_name, nspname)
        return columns[0] if columns else None

    def get_primary_key_columns(self, cursor, table_name, nspname):
        """Return a list of primary key columns for the given table."""
        for constraint in self.get_constraints(cursor, table_name, nspname).values():
            if constraint["primary_key"]:
                return constraint["columns"]
        return None

    def get_table_description(self, cursor, table_name, nspname):
        """
        Return a description of the table with the DB-API cursor.description
        interface.
        """
        # Query the pg_catalog tables as cursor.description does not reliably
        # return the nullable property and information_schema.columns does not
        # contain details of materialized views.
        cursor.execute(
            """
            SELECT
                a.attname AS column_name,
                NOT (a.attnotnull OR (t.typtype = 'd' AND t.typnotnull)) AS is_nullable,
                pg_get_expr(ad.adbin, ad.adrelid) AS column_default,
                CASE WHEN collname = 'default' THEN NULL ELSE collname END AS collation,
                a.attidentity != '' AS is_autofield,
                col_description(a.attrelid, a.attnum) AS column_comment
            FROM pg_attribute a
            LEFT JOIN pg_attrdef ad ON a.attrelid = ad.adrelid AND a.attnum = ad.adnum
            LEFT JOIN pg_collation co ON a.attcollation = co.oid
            JOIN pg_type t ON a.atttypid = t.oid
            JOIN pg_class c ON a.attrelid = c.oid
            JOIN pg_namespace n ON c.relnamespace = n.oid
            WHERE c.relkind IN ('f', 'm', 'p', 'r', 'v')
                AND c.relname = %s
                AND n.nspname NOT IN ('pg_catalog', 'pg_toast')
                AND n.nspname = ANY(current_schemas(false))
                AND n.nspname = %s
        """,
            [table_name, nspname],
        )
        field_map = {line[0]: line[1:] for line in cursor.fetchall()}

        _quote_name = self.connection.ops.quote_name
        cursor.execute(
            "SELECT * FROM %s LIMIT 1"
            % f"{_quote_name(nspname)}.{_quote_name(table_name)}"
        )
        return [
            FieldInfo(
                line.name,
                line.type_code,
                # display_size is always None on psycopg2.
                line.internal_size if line.display_size is None else line.display_size,
                line.internal_size,
                line.precision,
                line.scale,
                *field_map[line.name],
            )
            for line in cursor.description
        ]

    def get_attributes(self, cursor, table_name, nspname):
        attributes = dict()
        relations = self.get_relations(cursor, table_name, nspname)
        constraints = self.get_constraints(cursor, table_name, nspname)

        primary_key_column = self.get_primary_key_column(cursor, table_name, nspname)
        unique_columns = [
            c["columns"][0]
            for c in constraints.values()
            if c["unique"] and len(c["columns"]) == 1
        ]

        table_description = self.get_table_description(cursor, table_name, nspname)

        used_column_names = []
        column_to_field_name = {}

        for row in table_description:
            extra_params = {}
            column_name = row.name
            is_relation = column_name in relations

            att_name, params, notes = self.normalize_col_name(
                column_name, used_column_names, is_relation
            )
            extra_params.update(params)

            used_column_names.append(att_name)
            column_to_field_name[column_name] = column_name

            if primary_key_column is None and unique_columns:
                primary_key_column = unique_columns[0]

            if column_name == primary_key_column:
                extra_params["primary_key"] = True
            elif column_name in unique_columns:
                extra_params["unique"] = True

            if is_relation:
                ref_db_column, ref_db_table, ref_db_nspname = relations[column_name]
                if extra_params.pop("unique", False) or extra_params.get("primary_key"):
                    rel_type = "OneToOneField"

                else:
                    rel_type = "ForeignKey"
                    ref_pk_column = self.get_primary_key_column(
                        cursor, ref_db_table, ref_db_nspname
                    )
                    if ref_pk_column and ref_pk_column != ref_db_column:
                        extra_params["to_field"] = ref_db_column
                if (
                    relations[column_name][1] == table_name
                    and relations[column_name][2] == nspname
                ):
                    field_type, field_params, field_notes = self.get_field_type(row)
                    if (
                        field_type == "CharField"
                        and field_params.get("max_length", -1) < 0
                    ):
                        field_type = "TextField"
                    extra_params.update(field_params)
                    rel_type = field_type

                else:
                    rel_to = f"{relations[column_name][2]}_{relations[column_name][1]}"
                    extra_params["to"] = rel_to
                    extra_params["on_delete"] = models.DO_NOTHING
                field_type = rel_type
            else:
                field_type, field_params, field_notes = self.get_field_type(row)

                if field_type == "CharField" and field_params.get("max_length", -1) < 0:
                    field_type = "TextField"

                extra_params.update(field_params)

            if row.null_ok:
                extra_params["blank"] = True
                extra_params["null"] = True

            if extra_params.get("primary_key"):
                extra_params["blank"] = False
                extra_params["null"] = False
            extra_params["db_column"] = column_name

            field = models.__getattribute__(field_type)
            field = field(**extra_params)
            attributes[column_name] = field

        if "name" in attributes:
            attributes["__str__"] = lambda x: x.__getattribute__("name")

        return attributes

    def normalize_col_name(self, col_name, used_column_names, is_relation):
        return Command.normalize_col_name(
            Command, col_name, used_column_names, is_relation
        )

    def get_field_type(self, row):
        field_params = {}
        field_notes = []

        try:
            field_type = super().get_field_type(row.type_code, row)
        except KeyError:
            field_type = "TextField"
            field_notes.append("This field type is a guess.")

        # Add max_length for all CharFields.
        if field_type == "CharField" and row.display_size:
            if (size := int(row.display_size)) and size > 0:
                field_params["max_length"] = size

        if field_type in {"CharField", "TextField"} and row.collation:
            field_params["db_collation"] = row.collation

        if field_type == "DecimalField":
            if row.precision is None or row.scale is None:
                field_notes.append(
                    "max_digits and decimal_places have been guessed, as this "
                    "database handles decimal fields as float"
                )
                field_params["max_digits"] = (
                    row.precision if row.precision is not None else 10
                )
                field_params["decimal_places"] = (
                    row.scale if row.scale is not None else 5
                )
            else:
                field_params["max_digits"] = row.precision
                field_params["decimal_places"] = row.scale

        return field_type, field_params, field_notes



from django.test import TestCase
from django.test.utils import isolate_apps
from django.db import models

from django_schema_sprout.utils.dynamic_serializer import create_serializer


class DynamicSerializerTestCase(TestCase):
    @isolate_apps("test")
    def test_create_serializer(self):
        # Create a test model
        class TestModel(models.Model):
            class Meta:
                app_label = "test"

            username = models.CharField(max_length=100, primary_key=True)
            name = models.CharField(max_length=100)
            age = models.IntegerField()

        # # Create a serializer for the test model
        model_serializer = create_serializer(TestModel)

        # # Create an instance of the test model
        test_instance = TestModel(name="John Doe", age=30, username="johndoe")

        # # Create a serializer instance
        serializer = model_serializer(instance=test_instance)
        # # Assert that the serializer data matches the test instance data
        self.assertDictEqual(
            serializer.data, {"name": "John Doe", "age": 30, "username": "johndoe"}
        )



import django

django.setup()

TIME_ZONE = "UTC"
USE_TZ = True

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
    }
}

INSTALLED_APPS = (
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django_schema_sprout",
)



import tempfile

import pytest
from pytest_postgresql import factories
from django.conf import settings
from django.db import connections
from rest_framework.routers import DefaultRouter
from rest_framework.viewsets import ModelViewSet, ReadOnlyModelViewSet

from django_schema_sprout.schema_sprout import SchemaSprout
from django.db import ProgrammingError


# Using the factory to create a postgresql instance
socket_dir = tempfile.TemporaryDirectory()
postgresql_my_proc = factories.postgresql_proc(port=None, unixsocketdir=socket_dir.name)
postgresql_my = factories.postgresql("postgresql_my_proc")


@pytest.fixture(scope="function")
def database(postgresql_my):
    settings.DATABASES["django_test"] = {
        "ENGINE": "django.db.backends.postgresql_psycopg2",
    }
    with postgresql_my.cursor() as cursor:
        cursor.execute("CREATE SCHEMA foo;")
        cursor.execute("GRANT ALL ON SCHEMA foo TO postgres;")
        cursor.execute("GRANT USAGE ON SCHEMA foo TO postgres;")

        cursor.execute(
            """
        CREATE TABLE test (
            id serial,
            num FLOAT,
            data varchar,
            PRIMARY KEY(id, num)
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE foo.test (
            id serial,
            num FLOAT,
            data varchar,
            PRIMARY KEY(id, num)
        );"""
        )

        cursor.execute("CREATE VIEW test_view as SELECT * from test")

        cursor.execute(
            """
        CREATE TABLE foo.span (
            span_id TEXT PRIMARY KEY,
            span_text TEXT UNIQUE,
            span_num INTEGER
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE foo.eggs (
            eggs INTEGER PRIMARY KEY,
            eggs_text TEXT,
            span TEXT REFERENCES foo.span (span_id)
        );
        """
        )

        cursor.execute(
            """
            CREATE TABLE studies (
                nct_id varchar PRIMARY KEY,
                CONSTRAINT studies_nct_id_fkey FOREIGN KEY (nct_id)
                REFERENCES studies(nct_id)
            );
            """
        )
        cursor.execute('SET search_path = "$user", public, foo;')

    old_cursor = connections["django_test"].cursor
    connections["django_test"].cursor = postgresql_my.cursor

    yield postgresql_my

    connections["django_test"].cursor = old_cursor


@pytest.fixture(scope="function")
def schema_sprout():
    schema_sprout = SchemaSprout("django_test")
    yield schema_sprout
    SchemaSprout._instances = {}
    SchemaSprout._init = {}


def test_schema_sprout_init(database, schema_sprout):
    assert schema_sprout.connection == connections["django_test"]
    assert schema_sprout.views == {}
    assert schema_sprout.serializers == {}
    assert schema_sprout.models == {}
    assert isinstance(schema_sprout.router, DefaultRouter)


def _test_create_models(_schema_sprout, viewset_class):
    schema_sprout = _schema_sprout

    assert len(schema_sprout.models) == 5

    assert "foo_span" in schema_sprout.models
    assert "foo_eggs" in schema_sprout.models
    # Same tables in different schemas should be created with different models
    assert "public_test" in schema_sprout.models
    assert "foo_test" in schema_sprout.models
    assert "public_studies" in schema_sprout.models

    # Views should not be created as models
    assert "public_test_view" not in schema_sprout.models

    assert len(schema_sprout.serializers) == 5
    assert len(schema_sprout.views) == 5

    for url_path in [
        "foo/span",
        "foo/eggs",
        "public/test",
        "foo/test",
        "public/studies",
    ]:
        registered_correctly = False
        for registry in schema_sprout.router.registry:
            if registry[0] == url_path and issubclass(registry[1], viewset_class):
                registered_correctly = True

        assert registered_correctly


def test_schema_sprout_create_models(database, schema_sprout):
    schema_sprout.create_models()

    _test_create_models(schema_sprout, ModelViewSet)


def test_schema_sprout_create_models_read_only(database, schema_sprout):
    schema_sprout.create_models(readonly=True)
    _test_create_models(schema_sprout, ReadOnlyModelViewSet)


def test_schema_sprout_create_model(database, schema_sprout):
    schema_sprout.create_model("test", "public")

    assert len(schema_sprout.models) == 1
    assert "public_test" in schema_sprout.models

    assert schema_sprout.router.registry[0][0] == "public/test"
    assert issubclass(schema_sprout.router.registry[0][1], ModelViewSet)


def test_schema_sprout_create_same_model_two_times(database, schema_sprout):
    schema_sprout.create_model("test", "public")
    schema_sprout.create_model("test", "public")

    assert len(schema_sprout.models) == 1
    assert "public_test" in schema_sprout.models

    assert schema_sprout.router.registry[0][0] == "public/test"
    assert issubclass(schema_sprout.router.registry[0][1], ModelViewSet)


def test_schema_sprout_create_model_read_only(database, schema_sprout):
    schema_sprout.create_model("test", "public", readonly=True)

    assert len(schema_sprout.models) == 1
    assert "public_test" in schema_sprout.models

    assert schema_sprout.router.registry[0][0] == "public/test"
    assert issubclass(schema_sprout.router.registry[0][1], ReadOnlyModelViewSet)


def test_schema_sprout_create_model_table_permission_denied(
    database, schema_sprout, monkeypatch
):
    def mock_get_attributes(*args, **kwargs):
        raise ProgrammingError("access denied on table baz")

    monkeypatch.setattr(
        schema_sprout.db_inspector, "get_attributes", mock_get_attributes
    )

    schema_sprout.create_model("bar", "baz")

    assert len(schema_sprout.models) == 0
    assert "public_test" not in schema_sprout.models

    assert schema_sprout.router.registry == []



import tempfile

import pytest
from pytest_postgresql import factories
from django.conf import settings
from django.db import connections
from django.db.models import DO_NOTHING
from django.db.models.fields import (
    TextField,
    IntegerField,
    AutoField,
    CharField,
    FloatField,
)
from django.db.models.fields.related import ForeignKey

from django_schema_sprout.db_inspectors.postgres_inspector import (
    PostgresDBInspect,
    TableInfo,
    FieldInfo,
)

# Using the factory to create a postgresql instance
socket_dir = tempfile.TemporaryDirectory()
postgresql_my_proc = factories.postgresql_proc(port=None, unixsocketdir=socket_dir.name)
postgresql_my = factories.postgresql("postgresql_my_proc")


@pytest.fixture(scope="function")
def database(postgresql_my):
    settings.DATABASES["django_test"] = {
        "ENGINE": "django.db.backends.postgresql_psycopg2",
    }
    with postgresql_my.cursor() as cursor:
        cursor.execute("CREATE SCHEMA foo;")
        cursor.execute("CREATE SCHEMA bar;")
        cursor.execute("GRANT ALL ON SCHEMA foo TO postgres;")
        cursor.execute("GRANT USAGE ON SCHEMA foo TO postgres;")
        cursor.execute("GRANT ALL ON SCHEMA bar TO postgres;")
        cursor.execute("GRANT USAGE ON SCHEMA bar TO postgres;")

        cursor.execute(
            """
        CREATE TABLE test (
            id serial,
            num FLOAT,
            data varchar,
            PRIMARY KEY(id, num)
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE foo.span (
            span_id TEXT PRIMARY KEY,
            span_text TEXT UNIQUE,
            span_num INTEGER
        );
        """
        )
        cursor.execute(
            """
        CREATE TABLE foo.eggs (
            eggs INTEGER PRIMARY KEY,
            eggs_text TEXT,
            span TEXT REFERENCES foo.span (span_id)
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE bar.baz (
            id serial PRIMARY KEY,
            foo_span TEXT REFERENCES foo.span (span_id)
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE bar.az (
            id serial PRIMARY KEY,
            baz INTEGER REFERENCES bar.baz (id)
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE bar.ab (
            id serial PRIMARY KEY,
            az INTEGER REFERENCES bar.az (id)
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE studies (
            nct_id VARCHAR PRIMARY KEY,
            CONSTRAINT studies_nct_id_fkey FOREIGN KEY (nct_id)
            REFERENCES studies(nct_id)
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE simple (
            name VARCHAR
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE unique_without_primary (
            name VARCHAR UNIQUE
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE primary_with_uique (
            id INTEGER PRIMARY KEY,
            name TEXT UNIQUE
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE reference_to_unique (
            id INTEGER PRIMARY KEY,
            name TEXT REFERENCES primary_with_uique (name)
        );
        """
        )

        cursor.execute(
            """
        CREATE TABLE test_table (
            id SERIAL PRIMARY KEY,
            name VARCHAR(255) NOT NULL,
            age INT,
            email VARCHAR(255) UNIQUE,
            CONSTRAINT check_age CHECK (age >= 18)
        );

        CREATE INDEX idx_name ON test_table (name);
        CREATE INDEX idx_email ON test_table (email);
        """
        )

        cursor.execute('SET search_path = "$user", public, foo, bar;')

    yield postgresql_my


@pytest.mark.django_db
def test_get_table_list(database):
    inspector = PostgresDBInspect(connections["django_test"])
    with database.cursor() as cursor:
        expected_table_list = [
            TableInfo(name="span", type="t", comment=None, nspname="foo"),
            TableInfo(
                name="primary_with_uique", type="t", comment=None, nspname="public"
            ),
            TableInfo(name="simple", type="t", comment=None, nspname="public"),
            TableInfo(name="studies", type="t", comment=None, nspname="public"),
            TableInfo(name="test", type="t", comment=None, nspname="public"),
            TableInfo(name="test_table", type="t", comment=None, nspname="public"),
            TableInfo(
                name="unique_without_primary", type="t", comment=None, nspname="public"
            ),
            TableInfo(name="baz", type="t", comment=None, nspname="bar"),
            TableInfo(name="eggs", type="t", comment=None, nspname="foo"),
            TableInfo(
                name="reference_to_unique", type="t", comment=None, nspname="public"
            ),
            TableInfo(name="az", type="t", comment=None, nspname="bar"),
            TableInfo(name="ab", type="t", comment=None, nspname="bar"),
        ]
        table_list = inspector.get_table_list(cursor)

        assert table_list == expected_table_list


def test_get_relations(database):
    inspector = PostgresDBInspect(connections["django_test"])
    with database.cursor() as cursor:
        assert {} == inspector.get_relations(cursor, "span", "foo")
        assert {} == inspector.get_relations(cursor, "simple", "public")
        assert {"nct_id": ("nct_id", "studies", "public")} == inspector.get_relations(
            cursor, "studies", "public"
        )
        assert {} == inspector.get_relations(cursor, "test", "public")
        assert {} == inspector.get_relations(cursor, "baz", "bar")
        assert {"span": ("span_id", "span", "foo")} == inspector.get_relations(
            cursor, "eggs", "foo"
        )
        assert {"baz": ("id", "baz", "bar")} == inspector.get_relations(
            cursor, "az", "bar"
        )
        assert {"az": ("id", "az", "bar")} == inspector.get_relations(
            cursor, "ab", "bar"
        )
        assert {} == inspector.get_relations(cursor, "test_table", "public")


def test_get_constraints(database):
    inspector = PostgresDBInspect(connections["django_test"])
    with database.cursor() as cursor:
        assert inspector.get_constraints(cursor, "span", "foo") == {
            "span_pkey": {
                "columns": ["span_id"],
                "primary_key": True,
                "unique": True,
                "foreign_key": None,
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
            "span_span_text_key": {
                "columns": ["span_text"],
                "primary_key": False,
                "unique": True,
                "foreign_key": None,
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
        }
        assert inspector.get_constraints(cursor, "simple", "public") == {}
        assert inspector.get_constraints(cursor, "studies", "public") == {
            "studies_pkey": {
                "columns": ["nct_id"],
                "primary_key": True,
                "unique": True,
                "foreign_key": None,
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
            "studies_nct_id_fkey": {
                "columns": ["nct_id"],
                "primary_key": False,
                "unique": False,
                "foreign_key": ("studies", "nct_id"),
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
        }
        assert inspector.get_constraints(cursor, "test", "public") == {
            "test_pkey": {
                "columns": ["id", "num"],
                "primary_key": True,
                "unique": True,
                "foreign_key": None,
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            }
        }
        assert inspector.get_constraints(cursor, "baz", "bar") == {
            "baz_pkey": {
                "columns": ["id"],
                "primary_key": True,
                "unique": True,
                "foreign_key": None,
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
            "baz_foo_span_fkey": {
                "columns": ["foo_span"],
                "primary_key": False,
                "unique": False,
                "foreign_key": ("span", "span_id"),
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
        }
        assert inspector.get_constraints(cursor, "eggs", "foo") == {
            "eggs_pkey": {
                "columns": ["eggs"],
                "primary_key": True,
                "unique": True,
                "foreign_key": None,
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
            "eggs_span_fkey": {
                "columns": ["span"],
                "primary_key": False,
                "unique": False,
                "foreign_key": ("span", "span_id"),
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
        }
        assert inspector.get_constraints(cursor, "az", "bar") == {
            "az_pkey": {
                "columns": ["id"],
                "primary_key": True,
                "unique": True,
                "foreign_key": None,
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
            "az_baz_fkey": {
                "columns": ["baz"],
                "primary_key": False,
                "unique": False,
                "foreign_key": ("baz", "id"),
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
        }
        assert inspector.get_constraints(cursor, "ab", "bar") == {
            "ab_pkey": {
                "columns": ["id"],
                "primary_key": True,
                "unique": True,
                "foreign_key": None,
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
            "ab_az_fkey": {
                "columns": ["az"],
                "primary_key": False,
                "unique": False,
                "foreign_key": ("az", "id"),
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
        }
        assert inspector.get_constraints(cursor, "test_table", "public") == {
            "check_age": {
                "columns": ["age"],
                "primary_key": False,
                "unique": False,
                "foreign_key": None,
                "check": True,
                "index": False,
                "definition": None,
                "options": None,
            },
            "test_table_pkey": {
                "columns": ["id"],
                "primary_key": True,
                "unique": True,
                "foreign_key": None,
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
            "test_table_email_key": {
                "columns": ["email"],
                "primary_key": False,
                "unique": True,
                "foreign_key": None,
                "check": False,
                "index": False,
                "definition": None,
                "options": None,
            },
            "idx_email": {
                "columns": ["email"],
                "orders": ["ASC"],
                "primary_key": False,
                "unique": False,
                "foreign_key": None,
                "check": False,
                "index": True,
                "type": "idx",
                "definition": None,
                "options": None,
            },
            "idx_name": {
                "columns": ["name"],
                "orders": ["ASC"],
                "primary_key": False,
                "unique": False,
                "foreign_key": None,
                "check": False,
                "index": True,
                "type": "idx",
                "definition": None,
                "options": None,
            },
        }


def test_get_primary_key_column(database):
    inspector = PostgresDBInspect(connections["django_test"])
    with database.cursor() as cursor:
        assert inspector.get_primary_key_column(cursor, "span", "foo") == "span_id"
        assert inspector.get_primary_key_column(cursor, "simple", "public") is None
        assert inspector.get_primary_key_column(cursor, "studies", "public") == "nct_id"
        assert inspector.get_primary_key_column(cursor, "test", "public") == "id"
        assert inspector.get_primary_key_column(cursor, "baz", "bar") == "id"
        assert inspector.get_primary_key_column(cursor, "eggs", "foo") == "eggs"
        assert inspector.get_primary_key_column(cursor, "az", "bar") == "id"
        assert inspector.get_primary_key_column(cursor, "ab", "bar") == "id"
        assert inspector.get_primary_key_column(cursor, "test_table", "public") == "id"


def test_get_primary_key_columns(database):
    inspector = PostgresDBInspect(connections["django_test"])
    with database.cursor() as cursor:
        assert inspector.get_primary_key_columns(cursor, "span", "foo") == ["span_id"]
        assert inspector.get_primary_key_columns(cursor, "simple", "public") is None
        assert inspector.get_primary_key_columns(cursor, "studies", "public") == [
            "nct_id"
        ]
        assert inspector.get_primary_key_columns(cursor, "test", "public") == [
            "id",
            "num",
        ]
        assert inspector.get_primary_key_columns(cursor, "baz", "bar") == ["id"]
        assert inspector.get_primary_key_columns(cursor, "eggs", "foo") == ["eggs"]
        assert inspector.get_primary_key_columns(cursor, "az", "bar") == ["id"]
        assert inspector.get_primary_key_columns(cursor, "ab", "bar") == ["id"]
        assert inspector.get_primary_key_columns(cursor, "test_table", "public") == [
            "id"
        ]


def test_get_table_description(database):
    inspector = PostgresDBInspect(connections["django_test"])
    with database.cursor() as cursor:
        assert inspector.get_table_description(cursor, "span", "foo") == [
            FieldInfo(
                name="span_id",
                type_code=25,
                display_size=None,
                internal_size=None,
                precision=None,
                scale=None,
                null_ok=False,
                default=None,
                collation=None,
                is_autofield=False,
                comment=None,
            ),
            FieldInfo(
                name="span_text",
                type_code=25,
                display_size=None,
                internal_size=None,
                precision=None,
                scale=None,
                null_ok=True,
                default=None,
                collation=None,
                is_autofield=False,
                comment=None,
            ),
            FieldInfo(
                name="span_num",
                type_code=23,
                display_size=4,
                internal_size=4,
                precision=None,
                scale=None,
                null_ok=True,
                default=None,
                collation=None,
                is_autofield=False,
                comment=None,
            ),
        ]
        assert inspector.get_table_description(cursor, "simple", "public") == [
            FieldInfo(
                name="name",
                type_code=1043,
                display_size=None,
                internal_size=None,
                precision=None,
                scale=None,
                null_ok=True,
                default=None,
                collation=None,
                is_autofield=False,
                comment=None,
            )
        ]


def _test_attributes(expected, actual):
    assert expected.keys() == actual.keys()

    for key in expected:
        assert actual[key].__class__ == expected[key].pop("class")
        if "__fk_fields" in expected[key]:
            fk_fields = expected[key].pop("__fk_fields")
            actual[key].remote_field.on_delete = fk_fields["on_delete"]
            actual[key].remote_field.db_column = fk_fields["db_column"]
            actual[key].remote_model = fk_fields["to"]

        for attr in expected[key]:
            assert getattr(actual[key], attr) == expected[key][attr]


def test_get_attributes(database):
    inspector = PostgresDBInspect(connections["django_test"])
    with database.cursor() as cursor:
        expected_span_foo_attrs = {
            "span_id": {
                "primary_key": True,
                "blank": False,
                "null": False,
                "class": TextField,
            },
            "span_text": {
                "unique": True,
                "blank": True,
                "null": True,
                "class": TextField,
            },
            "span_num": {"class": IntegerField},
        }
        span_foo_attrs = inspector.get_attributes(cursor, "span", "foo")
        _test_attributes(expected_span_foo_attrs, span_foo_attrs)

        expected_simple_public_attrs = {
            "name": {"blank": True, "null": True, "class": TextField},
            "__str__": {"class": type(lambda: None)},
        }
        simple_public_attrs = inspector.get_attributes(cursor, "simple", "public")
        _test_attributes(expected_simple_public_attrs, simple_public_attrs)

        expected_studies_public_attrs = {
            "nct_id": {
                "primary_key": True,
                "blank": False,
                "null": False,
                "class": TextField,
            }
        }
        studies_public_attrs = inspector.get_attributes(cursor, "studies", "public")
        _test_attributes(expected_studies_public_attrs, studies_public_attrs)

        expected_test_public_attrs = {
            "id": {
                "primary_key": True,
                "blank": True,
                "null": False,
                "class": AutoField,
            },
            "num": {
                "primary_key": False,
                "blank": False,
                "null": False,
                "class": FloatField,
            },
            "data": {"blank": True, "null": True, "class": TextField},
        }
        test_public_attrs = inspector.get_attributes(cursor, "test", "public")
        _test_attributes(expected_test_public_attrs, test_public_attrs)

        expected_baz_bar_attrs = {
            "id": {
                "primary_key": True,
                "blank": True,
                "null": False,
                "class": AutoField,
            },
            "foo_span": {
                "primary_key": False,
                "blank": True,
                "null": True,
                "class": TextField,
            },
        }
        baz_bar_attrs = inspector.get_attributes(cursor, "baz", "bar")
        _test_attributes(expected_baz_bar_attrs, baz_bar_attrs)

        expected_eggs_foo_attrs = {
            "eggs": {
                "primary_key": True,
                "blank": False,
                "null": False,
                "class": IntegerField,
            },
            "eggs_text": {"blank": True, "null": True, "class": TextField},
            "span": {
                "primary_key": False,
                "blank": True,
                "null": True,
                "class": ForeignKey,
                "__fk_fields": {
                    "to": "foo_span",
                    "db_column": "span_id",
                    "on_delete": DO_NOTHING,
                },
            },
        }
        eggs_foo_attrs = inspector.get_attributes(cursor, "eggs", "foo")
        _test_attributes(expected_eggs_foo_attrs, eggs_foo_attrs)

        expected_az_bar_attrs = {
            "id": {
                "primary_key": True,
                "blank": True,
                "null": False,
                "class": AutoField,
            },
            "baz": {
                "primary_key": False,
                "blank": True,
                "null": True,
                "class": ForeignKey,
                "__fk_fields": {
                    "to": "baz",
                    "db_column": "id",
                    "on_delete": DO_NOTHING,
                },
            },
        }
        az_bar_attrs = inspector.get_attributes(cursor, "az", "bar")
        _test_attributes(expected_az_bar_attrs, az_bar_attrs)

        expected_ab_bar_attrs = {
            "id": {
                "primary_key": True,
                "blank": True,
                "null": False,
                "class": AutoField,
            },
            "az": {
                "primary_key": False,
                "blank": True,
                "null": True,
                "class": ForeignKey,
                "__fk_fields": {
                    "to": "az",
                    "db_column": "id",
                    "on_delete": DO_NOTHING,
                },
            },
        }
        ab_bar_attrs = inspector.get_attributes(cursor, "ab", "bar")
        _test_attributes(expected_ab_bar_attrs, ab_bar_attrs)

        expected_test_table_public_attrs = {
            "id": {
                "primary_key": True,
                "blank": True,
                "null": False,
                "class": AutoField,
            },
            "name": {"blank": False, "null": False, "class": CharField},
            "age": {"blank": True, "null": True, "class": IntegerField},
            "email": {
                "unique": True,
                "blank": True,
                "null": True,
                "class": CharField,
            },
            "__str__": {"class": type(lambda: None)},
        }
        test_table_public_attrs = inspector.get_attributes(
            cursor, "test_table", "public"
        )
        _test_attributes(expected_test_table_public_attrs, test_table_public_attrs)

        expected_unique_without_primary = {
            "name": {
                "primary_key": True,
                "blank": False,
                "null": False,
                "class": TextField,
            },
            "__str__": {"class": type(lambda: None)},
        }
        unique_without_primary_public_attrs = inspector.get_attributes(
            cursor, "unique_without_primary", "public"
        )
        _test_attributes(
            expected_unique_without_primary, unique_without_primary_public_attrs
        )

        expected_reference_to_unique = {
            "name": {
                "primary_key": False,
                "blank": True,
                "null": True,
                "class": ForeignKey,
                "__fk_fields": {
                    "to": "primary_with_uique",
                    "db_column": "name",
                    "on_delete": DO_NOTHING,
                },
            },
            "id": {"primary_key": True, "class": IntegerField},
            "__str__": {"class": type(lambda: None)},
        }
        actual_reference_to_unique = inspector.get_attributes(
            cursor, "reference_to_unique", "public"
        )
        _test_attributes(expected_reference_to_unique, actual_reference_to_unique)


def test_get_field_type():
    inspector = PostgresDBInspect(connections["django_test"])
    row = FieldInfo(
        name="wallet",
        type_code=1700,
        display_size=None,
        internal_size=None,
        precision=None,
        scale=None,
        null_ok=True,
        default=None,
        collation=None,
        is_autofield=False,
        comment=None,
    )

    field_type, field_params, field_notes = inspector.get_field_type(row)
    assert field_type == "DecimalField"
    assert field_params == {"max_digits": 10, "decimal_places": 5}
    assert field_notes == [
        "max_digits and decimal_places have been guessed, as this "
        "database handles decimal fields as float"
    ]

    row = FieldInfo(
        name="wallet",
        type_code=1700,
        display_size=None,
        internal_size=None,
        precision=5,
        scale=3,
        null_ok=True,
        default=None,
        collation=None,
        is_autofield=False,
        comment=None,
    )

    field_type, field_params, field_notes = inspector.get_field_type(row)
    assert field_type == "DecimalField"
    assert field_params == {"max_digits": 5, "decimal_places": 3}
    assert field_notes == []

    row = FieldInfo(
        name="wallet",
        type_code=1043,
        display_size=30,
        internal_size=None,
        precision=5,
        scale=3,
        null_ok=True,
        default=None,
        collation="C",
        is_autofield=False,
        comment=None,
    )

    field_type, field_params, field_notes = inspector.get_field_type(row)
    assert field_type == "CharField"
    assert field_params == {"max_length": 30, "db_collation": "C"}
    assert field_notes == []

    row = FieldInfo(
        name="wallet",
        type_code=666,
        display_size=30,
        internal_size=None,
        precision=5,
        scale=3,
        null_ok=True,
        default=None,
        collation="C",
        is_autofield=False,
        comment=None,
    )

    field_type, field_params, field_notes = inspector.get_field_type(row)
    assert field_type == "TextField"
    assert field_params == {"db_collation": "C"}
    assert field_notes == ["This field type is a guess."]



from django_schema_sprout.models import DynamicDbModel
from rest_framework import viewsets
from rest_framework import serializers


def create_view(
    model: DynamicDbModel,
    serializer: serializers.ModelSerializer,
    readonly: bool = False,
) -> type:
    """
    Create a view class for the given model and serializer.

    Args:
        serializer (serializers.ModelSerializer): The serializer class to use.
        model (DynamicDbModel): The model to create the view for.
        readonly (bool, optional): Whether the view should be read-only. Default False.

    Returns:
        type: The dynamically created view class.
    """
    # Implementation code goes here
    queryset = model.objects.all()
    serializer_class = serializer
    if readonly:
        viewset = viewsets.ReadOnlyModelViewSet
    else:
        viewset = viewsets.ModelViewSet
    return type(
        f"{model.__name__}ViewSet",
        (viewset,),
        {"queryset": queryset, "serializer_class": serializer_class},
    )



from django.db import models

from django_schema_sprout.apps import SchemaSproutAppConfig

app_label = SchemaSproutAppConfig.name


class DynamicDbModel(models.Model):
    class Meta:
        abstract = True
        app_label = app_label



class SchemaSproutDBRouter:
    def db_for_read(self, model, **hints):
        """reading model based on params"""
        if not hasattr(model, "Meta"):
            return None
        return getattr(model.Meta, "_schema_sprout_db", None)

    def db_for_write(self, model, **hints):
        """writing model based on params"""
        if not hasattr(model, "Meta"):
            return None
        return getattr(model.Meta, "_schema_sprout_db", None)



from django.test import TestCase
from django_schema_sprout.router import SchemaSproutDBRouter


class RouterTestCase(TestCase):
    def setUp(self):
        class ModelWithMeta:
            class Meta:
                pass

        class ModelWithSchemaSproutDB:
            class Meta:
                _schema_sprout_db = "test_db"

        self.router = SchemaSproutDBRouter()

        self.model_without_meta = type("ModelWithoutMeta", (), {})
        self.model_with_meta = ModelWithMeta
        self.model_with_schema_sprout_db = ModelWithSchemaSproutDB

    def test_db_for_write(self):
        # Test case when model does not have Meta attribute
        self.assertIsNone(self.router.db_for_write(self.model_without_meta))

        # Test case when model has Meta attribute but no _schema_sprout_db attribute
        self.assertIsNone(self.router.db_for_write(self.model_with_meta))

        # Test case when model has Meta attribute with _schema_sprout_db attribute
        self.assertEqual(
            self.router.db_for_write(self.model_with_schema_sprout_db), "test_db"
        )

    def test_db_for_read(self):
        # Test case when model does not have Meta attribute
        self.assertIsNone(self.router.db_for_read(self.model_without_meta))

        # Test case when model has Meta attribute but no _schema_sprout_db attribute
        self.assertIsNone(self.router.db_for_read(self.model_with_meta))

        # Test case when model has Meta attribute with _schema_sprout_db attribute
        self.assertEqual(
            self.router.db_for_read(self.model_with_schema_sprout_db), "test_db"
        )
